
--[[

name:   Antix LogWindow
vers:   1.0.1
desc:   A simple log window that you can append text to.
auth:   by Cliff Earl
date:   June 2016
legal:  Distribute/ modify as you like, but retain my credits, thanks

histo:  v1.0.1 - No longer subclassed from shape
                 Tidied up init/dispose code
--]]

LogWindow = Core.class(Sprite)

require "bit"

function LogWindow:init (options)
  local x, y, w, textColor, scale
  
  local props = { -- Default options
    lines           = 12,
    font            = Font.getDefault(),
    x               = 16,
    y               = 16,
    w               = 300,
    textColor       = 0xeeeef0,
    scale           = 1,
    width            = 512,
  }
  
  if options then -- overlay any specified options
    for key, value in pairs(options) do
      props[key]= value
    end
  end
  
  self.lines        = props.lines
  self.font         = props.font
  x                 = props.x
  y                 = props.y
  self.width        = props.width
  self.textColor    = props.textColor
  scale             = props.scale

  local lineHeight = self.font:getLineHeight()
  local h = self.lines * lineHeight + 16
  self:setScale(scale)
  self:setPosition(x, y)

  for i = 1, self.lines do -- Create blank text lines
    local text = TextField.new(self.font, "")
    text:setPosition(8, (i * lineHeight) + 8)
    text:setTextColor(self.textColor)
    self:addChild(text)
  end
  
  stage:addChild(self)
end

-- Attempt to cleanup our mess :D
function LogWindow:shutDown()
  self:clear()
  for i = self:getNumChildren(), 1, -1 do -- Remove in reverse order
    local text = self:getChildAt(i)
    self:removeChildAt(i)
    text = nil
  end

  self.lines = nil
  self.font = nil
  self.width = nil
  self.textColor = nil
end

-- Attach LogWindow to new parent
function LogWindow:attach(newParent)
  self:detach()
  newParent:addChild(self)
end

-- Detach LogWindow from current parent
function LogWindow:detach()
  self:removeFromParent()
end

-- Show/Hide
function LogWindow:show(show)
  self:setVisible(show)
end

-- Bring to front of all others sprites attached to parent
function LogWindow:toFront()
  self:getParent():addChild(self)
  self:setVisible(true)
end

-- Clear all lines of text`
function LogWindow:clear()
  for i = 1, self:getNumChildren() do
    self:getChildAt(i):setText("")
  end
end

-- Append new text to end of log, scrolling existing lines up  
function LogWindow:append(text, color)
  local r, g, b = self:colorToRGB(color or self.textColor)
  
  local newLines = self:split(text, self.width - 16)

  if #newLines >= self.lines then
    local j = #newLines - self.lines -- Skip extra lines and set all lines
    for i = 1, self.lines do
      self:getChildAt(i):setText(newLines[i + j])
      self:getChildAt(i):setColorTransform(r, g, b)
    end
    
  else
    local j = 1 -- SCROLL BLOCK
    for i = #newLines, self.lines -1 do
      local source = self:getChildAt(i + 1)
      local dest = self:getChildAt(j)
      dest:setText(source:getText())
      dest:setColorTransform(source:getColorTransform())
      j = j + 1
    end
    
    j = 1 -- SET NEW TEXT
    for i = self.lines - #newLines + 1, self.lines do
      self:getChildAt(i):setText(newLines[j])
      self:getChildAt(i):setColorTransform(r, g, b)
      j = j + 1
    end
  end
end

-- Convert RGB to R,G,B saturation
function LogWindow:colorToRGB(color)
  local r = 1/255 * bit.band(bit.ror(color, 16), 255)
  local g = 1/255 * bit.band(bit.ror(color, 8), 255)
  local b = 1/255 * bit.band(color, 255)

  return r, g, b
end

-- Split text into multiple lines
function LogWindow:split(text, maxWidth)
  local insert = table.insert
  
	local function split(input, char) -- SPLIT FUNCTION FROM STACK EXCHANGE http://bit.ly/1iXGtNN
		local match = string.gmatch
		local pattern = '([^'..char..']+)' -- OPTIMIZED SLIGHTLY BY ANTIX
		local output = {}
		for str in match(input, pattern) do -- MORE OPTIMIZATIONS AFTER FULLY READING THREAD
			output[#output + 1] = str -- THIS IS FASTER THAN TABLE.INSERT
		end
		return output
	end

  local font = self.font
  
	local paragraphs = split(text, "|") -- SPLIT TEXT INTO PARAGRAPHS

	local lines = {}
	
	for p = 1, #paragraphs do

		if #paragraphs[p] == 1 then -- A SPECIAL CASE FOR LINEBREAKS
			insert(lines, " ")
		else
			local words = split(paragraphs[p], " ") -- SPLIT PARAGRAPH INTO WORDS
			local word = 1
			local line = ""
			local lineWidth = 0
			local done = false
			
			repeat
				local wordWidth = font:getAdvanceX(words[word] .. " ") -- GET PIXEL WIDTH OF NEXT WORD
				if lineWidth + wordWidth >= maxWidth then -- IF ABOUT TO EXCEED MAXIMUM WIDTH, ADD LINE TO LINES AND START A NEW LINE
					insert(lines, line)
					line = ""
					lineWidth = 0
				end
				line = line .. " " .. words[word] -- ADD WORD TO LINE, PLUS SPACE CHAR THAT WAS STRIPPED IN SPLIT FUNCTION
				lineWidth = lineWidth + wordWidth 
				word = word + 1
				if word > #words then -- IF ALL WORDS PROCESSED, ADD LAST LINE TO LINES AND SET FLAG TO BEGIN NEXT PARAGRAPH
					insert(lines, line)
					done = true
				end
			until done
    line = nil
    words = nil
		end
	end
	paragraphs = nil
	
	return lines
end
